"""Cython functions for grid calculations"""

# MIT License

# Copyright (c) 2025 Guillaume Rongier

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


cimport cython
from cython.parallel cimport prange

# https://cython.readthedocs.io/en/stable/src/userguide/fusedtypes.html
ctypedef fused float_or_int:
    cython.integral
    long long
    cython.floating


################################################################################
# Functions


@cython.boundscheck(False)
@cython.wraparound(False)
def _calc_mult_flux_div_at_node(
    shape,
    xy_spacing,
    const float_or_int[:, :] value_at_link,
    cython.floating[:, :] out,
):
    cdef int n_rows = shape[0]
    cdef int n_cols = shape[1]
    cdef int n_classes = value_at_link.shape[1]
    cdef double dx = xy_spacing[0]
    cdef double dy = xy_spacing[1]
    cdef int links_per_row = 2 * n_cols - 1
    cdef double inv_area_of_cell = 1.0 / (dx * dy)
    cdef int row, col, clas
    cdef int node, link

    for row in prange(1, n_rows - 1, nogil=True, schedule="static"):
        node = row * n_cols
        link = row * links_per_row

        for col in range(1, n_cols - 1):
            for clas in range(n_classes):
                out[node + col, clas] = (
                    dy * (value_at_link[link + 1, clas] - value_at_link[link, clas])
                    + dx * (value_at_link[link + n_cols, clas] - value_at_link[link - n_cols + 1, clas])
                ) * inv_area_of_cell
            link = link + 1



from .sea_level import SeaLevelCalculator
from .sea_level import CyclicSeaLevelCalculator
from .source import SourceProducer
from .weathering import BedrockWeatherer
from .diffusion import GravityDrivenDiffuser
from .diffusion import GravityDrivenRouter
from .stream_power import WaterDrivenDiffuser
from .stream_power import WaterDrivenRouter
from .stream_power import FluxDrivenRouter
from .landsliding import SimpleSedimentLandslider
from .compaction import SedimentCompactor

COMPONENTS = [
    SeaLevelCalculator,
    SourceProducer,
    CyclicSeaLevelCalculator,
    BedrockWeatherer,
    GravityDrivenDiffuser,
    GravityDrivenRouter,
    WaterDrivenDiffuser,
    WaterDrivenRouter,
    FluxDrivenRouter,
    SimpleSedimentLandslider,
    SedimentCompactor,
]

__all__ = [cls.__name__ for cls in COMPONENTS]



"""Source producer"""

# MIT License

# Copyright (c) 2025 Guillaume Rongier

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


from functools import partial
import numpy as np
from scipy.spatial import KDTree
from scipy.interpolate import CubicSpline, PchipInterpolator, make_interp_spline
from landlab import Component


################################################################################
# Component


class SourceProducer(Component):
    """Produce water and sediments from sources based on interpolating from
    control points.

    Negative fluxes produced by interpolation are truncated to zero.
    """

    _name = "SourceProducer"

    _unit_agnostic = True

    _info = {
        "water__unit_flux_in": {
            "dtype": float,
            "intent": "out",
            "optional": False,
            "units": "m/s",
            "mapping": "node",
            "doc": "External volume water per area per time input to each node (e.g., rainfall rate)",
        },
        "sediment__unit_flux_in": {
            "dtype": float,
            "intent": "out",
            "optional": False,
            "units": "m3/s",
            "mapping": "node",
            "doc": "Sediment flux as boundary condition",
        },
    }

    def __init__(
        self,
        grid,
        source_xy,
        water_source_time,
        water_source_influx,
        sediment_source_time,
        sediment_source_influx,
        interpolation="linear",
    ):
        """
        Parameters
        ----------
        grid : ModelGrid
            A grid.
        source_xy : array of shape (n_source, 2) or (2,)
            The x and y coordinates of the source(s). When two sources end up in
            the same cell, their contributions are added.
        water_source_time : array of shape (n_source, n_time) or (n_time,)
            The time at each control point for each source for the water influx.
        water_source_influx : array of shape (n_source, n_time) or (n_time,)
            The water influx at each control point for each source.
        sediment_source_time : array of shape (n_source, n_time) or (n_time,)
            The time at each control point for each source for the sediment influx.
        sediment_source_influx : array of shape (n_source, n_time, n_class), (n_time, n_class), or (n_time,)
            The sediment influx at each control point for each source and each
            sediment class.
        interpolation : str
            The interpolation method, which can be:
                - 'linear' for linear interpolation (the default).
                - 'quadratic' for spline interpolation of second order.
                - 'cubic' for spline interpolation of third order.
                - 'monotonic' for monotonic spline interpolation of third order.
        """
        super().__init__(grid)

        # Parameters
        source_xy = np.asarray(source_xy)
        if source_xy.ndim == 1:
            source_xy = source_xy[np.newaxis]
        water_source_influx = np.asarray(water_source_influx)
        if water_source_influx.ndim == 1:
            water_source_influx = np.tile(water_source_influx, (len(source_xy), 1))
        water_source_time = np.asarray(water_source_time)
        if water_source_time.ndim == 1:
            water_source_time = np.broadcast_to(
                water_source_time, water_source_influx.shape
            )
        sediment_source_influx = np.asarray(sediment_source_influx)
        if sediment_source_influx.ndim == 1:
            sediment_source_influx = np.tile(
                sediment_source_influx, (len(source_xy), 1)
            )
        if sediment_source_influx.ndim == 2:
            sediment_source_influx = sediment_source_influx[..., np.newaxis]
        sediment_source_time = np.asarray(sediment_source_time)
        if sediment_source_time.ndim == 1:
            sediment_source_time = np.tile(sediment_source_time, (len(source_xy), 1))
        self._time = 0.0

        # Fields
        if "water__unit_flux_in" not in grid.at_node.keys():
            self._water_influx = grid.add_zeros(
                "water__unit_flux_in", at="node", clobber=True
            )
        else:
            self._water_influx = grid.at_node["water__unit_flux_in"]
        if "sediment__unit_flux_in" not in grid.at_node.keys():
            self._sediment_influx = grid.add_field(
                "sediment__unit_flux_in",
                np.zeros((grid.number_of_nodes, sediment_source_influx.shape[2])),
                clobber=True,
            )
        else:
            self._sediment_influx = grid.at_node["sediment__unit_flux_in"]
        if self._sediment_influx.ndim == 1:
            self._sediment_influx = self._sediment_influx[:, np.newaxis]

        # Source location in the grid
        kdtree = KDTree(grid.xy_of_cell)
        _, self._source_idx = kdtree.query(source_xy)
        self._source_idx = grid.core_nodes[self._source_idx]

        # Interpolators
        self._water_interpolators = []
        for i in range(len(water_source_time)):
            self._water_interpolators.append(
                self._build_interpolator(
                    water_source_time[i], water_source_influx[i], interpolation
                )
            )
        self._sediment_interpolators = []
        for i in range(len(sediment_source_time)):
            _interpolators = []
            for j in range(sediment_source_influx.shape[2]):
                _interpolators.append(
                    self._build_interpolator(
                        sediment_source_time[i],
                        sediment_source_influx[i, :, j],
                        interpolation,
                    )
                )
            self._sediment_interpolators.append(_interpolators)

    def _build_interpolator(self, x, y, interpolation):
        """Builds an interpolator based on a given interpolation technique."""
        if interpolation == "linear":
            return partial(np.interp, xp=x, fp=y)
        elif interpolation == "quadratic":
            return make_interp_spline(x, y, k=2)
        elif interpolation == "cubic":
            return CubicSpline(x, y)
        elif interpolation == "monotonic":
            return PchipInterpolator(x, y)

    def run_one_step(self, dt):
        """Run the producer for one timestep, dt.

        Parameters
        ----------
        dt : float (time)
            The imposed timestep.
        """
        self._time += dt

        self._water_influx[self._source_idx] = 0.0
        for i, interpolator in enumerate(self._water_interpolators):
            value = interpolator(self._time)
            if value > 0.:
                self._water_influx[self._source_idx[i]] += value
        self._sediment_influx[self._source_idx] = 0.0
        for i, interpolators in enumerate(self._sediment_interpolators):
            for j, interpolator in enumerate(interpolators):
                value = interpolator(self._time)
                if value > 0.:
                    self._sediment_influx[self._source_idx[i], j] += value



"""Plotting layers"""

# MIT License

# Copyright (c) 2025 Guillaume Rongier

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


import warnings
import numpy as np

from ..utils import reshape_to_match
from .cfuncs import mask_layer


################################################################################
# Base functions


def _get_variable_values(grid, var):
    """
    Gets the values for a variable of a StackedLayers.
    """
    if isinstance(var, np.ndarray):
        if len(var) == grid.number_of_nodes:
            return var[grid.core_nodes]
        elif var.ndim > 1 and var.shape[1] == grid.number_of_nodes:
            return var[:, grid.core_nodes]
        else:
            return var
    elif var == "dz" or var == "thickness":
        return grid.stacked_layers["_dz"]
    elif var == "composition":
        return grid.stacked_layers.composition
    elif var == "most_frequent_class" or var == "most_frequent":
        return grid.stacked_layers.most_frequent_class
    else:
        return grid.stacked_layers[var]


def _select_sublayers(layers, i, axis=2, dz=None, indices=None):
    """
    Gets the values of a subselection of a stack of layers.
    """
    if layers.ndim == 1 or axis >= layers.ndim:
        return layers
    elif isinstance(i, int):
        slices = tuple(slice(None) if j != axis else i for j in range(layers.ndim))
        return layers[slices]
    elif callable(i):
        return i(layers, axis=axis)
    elif i == "middle":
        slices = tuple(
            slice(None) if j != axis else layers.shape[axis] // 2
            for j in range(layers.ndim)
        )
        return layers[slices]
    elif i == "weighted_mean" and dz is not None:
        thickness = np.sum(dz, axis=axis, keepdims=True)
        composition = np.zeros_like(dz)
        np.divide(dz, thickness, out=composition, where=thickness > 0.0)
        return np.sum(layers * composition, axis=axis)
    elif i == "top" and indices is not None:
        return np.take_along_axis(
            layers, reshape_to_match(indices, layers.shape), axis=axis
        )[0]
    else:
        return None


def _get_layer_values(grid, var, i_layer, i_class):
    """
    Gets the values of a layer.
    """
    if i_class == "weighted_mean":
        dz = _select_sublayers(
            _get_variable_values(grid, "_dz"),
            i_layer,
            axis=0,
            indices=grid.stacked_layers.surface_index,
        )
    else:
        dz = None

    return _select_sublayers(
        _select_sublayers(
            _get_variable_values(grid, var),
            i_layer,
            axis=0,
            indices=grid.stacked_layers.surface_index,
        ),
        i_class,
        axis=1,
        dz=dz,
    )


def _get_layer_elevation(grid):
    """
    Gets the elevation of a StackedLayers.
    """
    z = grid.at_node["topographic__elevation"][grid.core_nodes] - grid.stacked_layers.z

    return np.concatenate(
        [z, grid.at_node["topographic__elevation"][grid.core_nodes][np.newaxis]]
    )


def _plot_horizontal_slice(
    ax, grid, var, i_layer, i_class, mask_value, vmin, vmax, shading, **kwargs
):
    """
    Plots a horizontal slice through a StackedLayers on a RasterModelGrid.
    """
    x = grid.x_of_node[grid.core_nodes].reshape(grid.cell_grid_shape)
    y = grid.y_of_node[grid.core_nodes].reshape(grid.cell_grid_shape)
    values = _get_layer_values(grid, var, i_layer, i_class).reshape(
        grid.cell_grid_shape
    )
    if mask_value is not None:
        values[values == mask_value] = np.nan

    vmin = None if "norm" in kwargs else vmin
    vmax = None if "norm" in kwargs else vmax

    return ax.pcolormesh(x, y, values, vmin=vmin, vmax=vmax, shading=shading, **kwargs)


def _plot_vertical_slice(
    ax,
    grid,
    var,
    i_x,
    i_y,
    i_class,
    mask_wedges,
    mask_null_layers,
    mask_value,
    vmin,
    vmax,
    shading,
    **kwargs,
):
    """
    Plots a vertical slice through a StackedLayers on a RasterModelGrid.
    """
    x = grid.x_of_node[grid.core_nodes].reshape(grid.cell_grid_shape)
    y = grid.y_of_node[grid.core_nodes].reshape(grid.cell_grid_shape)
    z = _get_layer_elevation(grid).reshape(-1, *grid.cell_grid_shape)
    _dz = _get_variable_values(grid, "dz")
    dz = _select_sublayers(_dz, np.sum, axis=2).reshape(-1, *grid.cell_grid_shape)
    if (
        isinstance(var, np.ndarray) == False
        and (var == "_dz" or var == "dz" or var == "thickness")
        and i_class == np.sum
    ):
        values = dz
    else:
        values = _select_sublayers(
            _get_variable_values(grid, var), i_class, axis=2, dz=_dz
        ).reshape(-1, *grid.cell_grid_shape)
    if mask_value is not None:
        values[values == mask_value] = np.nan

    if i_x is None:
        i_x = slice(None)
        if i_y == "middle":
            i_y = grid.cell_grid_shape[0] // 2
        coords = np.tile(x[i_y, i_x], (2, 1))
    elif i_y is None:
        i_y = slice(None)
        if i_x == "middle":
            i_x = grid.cell_grid_shape[1] // 2
        coords = np.tile(y[i_y, i_x], (2, 1))

    vmin = (
        np.nanmin(values[..., i_y, i_x])
        if vmin is None and "norm" not in kwargs
        else vmin
    )
    vmax = (
        np.nanmax(values[..., i_y, i_x])
        if vmax is None and "norm" not in kwargs
        else vmax
    )

    c = []
    for i in range(len(values)):
        layer = values[i, i_y, i_x].astype(float)
        mask_layer(layer, dz[i, i_y, i_x], mask_wedges, mask_null_layers)
        ci = ax.pcolormesh(
            coords,
            z[i : i + 2, i_y, i_x],
            np.tile(layer, (2, 1)),
            shading=shading,
            vmin=vmin,
            vmax=vmax,
            **kwargs,
        )
        c.append(ci)

    return c


################################################################################
# Plot


def plot_layers(
    ax,
    grid,
    var,
    i_x="middle",
    i_y=None,
    i_layer=None,
    i_class=None,
    mask_wedges=False,
    mask_null_layers=True,
    mask_value=None,
    vmin=None,
    vmax=None,
    shading="gouraud",
    **kwargs,
):
    """Plot a horizontal or vertical slice through a stack of layers on a raster
    grid.

    Visualization follows the tie-centered approach described by Tetzlaff (2023);
    channel erosion and deposition is not properly handled in the vertical cross-
    sections (see figure 12 of Tetzlaff (2023)).

    Parameters
    ----------
    ax : matplotlib axes object
        The axis of a figure on which to plot the layer(s).
    grid : RasterModelGrid
        The grid containing the StackedLayers to plot.
    var : str or array-like
        The variable to plot as color, which can be on the StackedLayers or as
        a separate array. The keywords 'dz' or 'thickness' can be used to plot
        the layer thickness; the keyword 'composition' can be used to plot the
        proportion of each lithology; the keywords 'most_frequent_class' or
        'most_frequent' can be to plot the most frequent lithology.
    i_x : int, str, or callable, optional
        The index at which to slice vertically the layers along the grid's x
        axis. The keyword 'middle' will slice at the middle of the axis. A
        callable will merge the all the slices along the axis together (e.g.,
        numpy.mean or numpy.sum). A slice can only be made along a single axis,
        so `i_x` is incompatible with `i_y` and `i_layer`.
    i_y : int, str, or callable, optional
        The index at which to slice vertically the layers along the grid's y
        axis. The keyword 'middle' will slice at the middle of the axis. A
        callable will merge the all the slices along the axis together (e.g.,
        numpy.mean or numpy.sum). A slice can only be made along a single axis,
        so `i_y` is incompatible with `i_x` and `i_layer`.
    i_layer : int, str, or callable, optional
        The index at which to slice horizontally the layers, i.e., to plot a
        given layer in map view. The keyword 'middle' will slice at the middle
        of the axis; the keyword 'top' will show the values at the topographic
        surface. A callable will merge the all the slices along the axis together
        (e.g., numpy.mean or numpy.sum). A slice can only be made  along a single
        axis, so `i_layer` is incompatible with `i_x` and `i_y`.
    i_class : int or callable, optional
        The index to select the lithology to plot. A callable will merge the all
        the lithologies together (e.g., numpy.mean or numpy.sum). Using
        'weighted_mean' will average the lithologies weighted by their respective
        thicknesses.
    mask_wedges : bool, optional
        If True, wedges, where a layer pinches out to a null thickness, are
        displayed with the value of the non-null node. This is useful when
        displaying the composition of a given class for instance, and avoid
        perturbing the visualization with a pointless null composition.
    mask_null_layers : bool, optional
        If True, nodes of null thickness that are surrounded with null-thickness
        nodes are turned to NaN. This avoid perturbing the display of the
        stratigraphy, where null-thickness layers can still be visible.
    mask_value : float, optional
        A value of the variable being plotted to mask, i.e., to turn to NaN.
    vmin : float, optional
        The minimum value to use when plotting the variable. If None, uses the
        minimum value of the variable.
    vmax : float, optional
        The maximum value to use when plotting the variable. If None, uses the
        maximum value of the variable.
    shading : str
        The fill style for the quadrilateral, which can be 'flat', 'nearest',
        'gouraud', or 'auto'. By default, it reproduces the tie-centered
        scheme of Tetzlaff (2023) by using 'gouraud'.
    **kwargs : `matplotlib.axes.Axes.pcolormesh` properties
        Other keyword arguments are pcolormesh properties, see
        `matplotlib.axes.Axes.pcolormesh` for a list of valid properties.

    Returns
    -------
    c : matplotlib.collections.QuadMesh or list
        The mesh(es) that constitute(s) the plot.

    Reference
    ---------
    Tetzlaff, D. (2023)
        Stratigraphic forward modeling software package for research and education
        https://arxiv.org/abs/2302.05272
    """
    if i_x is not None and i_y is not None:
        warnings.warn(
            "Both `i_x` and `i_y` are not None, but a slice can only be made along a single axis. `i_y` will be turned to None."
        )
        i_y = None
    if i_x is not None and i_layer is not None:
        warnings.warn(
            "Both `i_x` and `i_layer` are not None, but a slice can only be made along a single axis. `i_layer` will be turned to None."
        )
        i_layer = None
    if i_y is not None and i_layer is not None:
        warnings.warn(
            "Both `i_y` and `i_layer` are not None, but a slice can only be made along a single axis. `i_layer` will be turned to None."
        )
        i_layer = None

    if i_x is not None or i_y is not None:
        return _plot_vertical_slice(
            ax,
            grid,
            var,
            i_x,
            i_y,
            i_class,
            mask_wedges,
            mask_null_layers,
            mask_value,
            vmin,
            vmax,
            shading,
            **kwargs,
        )
    elif i_layer is not None:
        return _plot_horizontal_slice(
            ax, grid, var, i_layer, i_class, mask_value, vmin, vmax, shading, **kwargs
        )
    else:
        raise Exception("At least one of `i_x`, `i_y`, or `i_layer` must not be None.")


class RasterModelGridLayerPlotterMixIn:
    """MixIn that provides layer plotting functionality to a raster grid.

    Inhert from this class to provide a ModelDataFields object with the method
    function, ``plot_layers``, that plots a slice through a StackedLayers.
    """

    def plot_layers(
        self,
        ax,
        var,
        i_x="middle",
        i_y=None,
        i_layer=None,
        i_class=None,
        mask_wedges=False,
        mask_null_layers=True,
        mask_value=None,
        vmin=None,
        vmax=None,
        shading="gouraud",
        **kwargs,
    ):
        """Plot a horizontal or vertical slice through a StackedLayers. This is
        a wrapper for `plot.plot_layers`, and its uses the same keywords.

        Visualization follows the tie-centered approach described by Tetzlaff (2023);
        channel erosion and deposition is not properly handled in the vertical cross-
        sections (see figure 12 of Tetzlaff (2023)).

        Parameters
        ----------
        ax : matplotlib axes object
            The axis of a figure on which to plot the layer(s).
        grid : RasterModelGrid
            The grid containing the StackedLayers to plot.
        var : str or array-like
            The variable to plot as color, which can be on the StackedLayers or as
            a separate array. The keywords 'dz' or 'thickness' can be used to plot
            the layer thickness; the keyword 'composition' can be used to plot the
            proportion of each lithology; the keywords 'most_frequent_class' or
            'most_frequent' can be to plot the most frequent lithology.
        i_x : int, str, or callable, optional
            The index at which to slice vertically the layers along the grid's x
            axis. The keyword 'middle' will slice at the middle of the axis. A
            callable will merge the all the slices along the axis together (e.g.,
            numpy.mean or numpy.sum). A slice can only be made along a single axis,
            so `i_x` is incompatible with `i_y` and `i_layer`.
        i_y : int, str, or callable, optional
            The index at which to slice vertically the layers along the grid's y
            axis. The keyword 'middle' will slice at the middle of the axis. A
            callable will merge the all the slices along the axis together (e.g.,
            numpy.mean or numpy.sum). A slice can only be made along a single axis,
            so `i_y` is incompatible with `i_x` and `i_layer`.
        i_layer : int, str, or callable, optional
            The index at which to slice horizontally the layers, i.e., to plot a
            given layer in map view. The keyword 'middle' will slice at the middle
            of the axis; the keyword 'top' will show the values at the topographic
            surface. A callable will merge the all the slices along the axis together
            (e.g., numpy.mean or numpy.sum). A slice can only be made  along a single
            axis, so `i_layer` is incompatible with `i_x` and `i_y`.
        i_class : int or callable, optional
            The index to select the lithology to plot. A callable will merge the all
            the lithologies together (e.g., numpy.mean or numpy.sum). Using
            'weighted_mean' will average the lithologies weighted by their respective
            thicknesses.
        mask_wedges : bool, optional
            If True, wedges, where a layer pinches out to a null thickness, are
            displayed with the value of the non-null node. This is useful when
            displaying the composition of a given class for instance, and avoid
            perturbing the visualization with a pointless null composition.
        mask_null_layers : bool, optional
            If True, nodes of null thickness that are surrounded with null-thickness
            nodes are turned to NaN. This avoid perturbing the display of the
            stratigraphy, where null-thickness layers can still be visible.
        mask_value : float, optional
            A value of the variable being plotted to mask, i.e., to turn to NaN.
        vmin : float, optional
            The minimum value to use when plotting the variable. If None, uses the
            minimum value of the variable.
        vmax : float, optional
            The maximum value to use when plotting the variable. If None, uses the
            maximum value of the variable.
        shading : str
            The fill style for the quadrilateral, which can be 'flat', 'nearest',
            'gouraud', or 'auto'. By default, it reproduces the tie-centered
            scheme of Tetzlaff (2023) by using 'gouraud'.
        **kwargs : `matplotlib.axes.Axes.pcolormesh` properties
            Other keyword arguments are pcolormesh properties, see
            `matplotlib.axes.Axes.pcolormesh` for a list of valid properties.

        Returns
        -------
        c : matplotlib.collections.QuadMesh or list
            Mesh(es) that constitute(s) the plot.

        Reference
        ---------
        Tetzlaff, D. (2023)
            Stratigraphic forward modeling software package for research and education
            https://arxiv.org/abs/2302.05272

        See Also
        --------
        stratigrapy.plot.plot_layers
        """
        return plot_layers(
            ax,
            self,
            var,
            i_x,
            i_y,
            i_layer,
            i_class,
            mask_wedges,
            mask_null_layers,
            mask_value,
            vmin,
            vmax,
            shading,
            **kwargs,
        )


def extract_tie_centered_layers(
    grid,
    var,
    i_class=None,
    axis=2,
    mask_wedges=False,
    mask_null_layers=False,
    mask_value=None,
    fill_nan=False,
):
    """Extract the layers coordinates and values of a variable for plotting, in
    particular with PyVista.

    Visualization follows the tie-centered approach described by Tetzlaff (2023);
    channel erosion and deposition is not properly handled in the vertical cross-
    sections (see figure 12 of Tetzlaff (2023)).

    Parameters
    ----------
    grid : RasterModelGrid
        The grid containing the StackedLayers to plot.
    var : str or array-like
        The variable to plot as color, which can be on the StackedLayers or as
        a separate array. The keywords 'dz' or 'thickness' can be used to plot
        the layer thickness; the keyword 'composition' can be used to plot the
        proportion of each lithology.
    i_class : int or callable, optional
        The index to select the lithology to plot. A callable will merge the all
        the lithologies together (e.g., numpy.mean or numpy.sum). Using
        'weighted_mean' will average the lithologies weighted by their respective
        thicknesses.
    axis : int
        The axis along which to post-process the layers to manage pinching-out
        cells. By default, the x-axis is used.
    mask_wedges : bool, optional
        If True, wedges, where a layer pinches out to a null thickness, are
        displayed with the value of the non-null node. This is useful when
        displaying the composition of a given class for instance, and avoid
        perturbing the visualization with a pointless null composition.
    mask_null_layers : bool, optional
        If True, nodes of null thickness that are surrounded with null-thickness
        nodes are turned to NaN. This avoid perturbing the display of the
        stratigraphy, where null-thickness layers can still be visible.
    mask_value : float, optional
        A value of the variable being plotted to mask, i.e., to turn to NaN.
    fill_nan: bool
        If True, fills the NaN values of a layer with the values of the layer
        below.

    Returns
    -------
    x : ndarray
        The coordinates along the x-axis of the 3D grid making the layers.
    y : ndarray
        The coordinates along the y-axis of the 3D grid making the layers.
    z : ndarray
        The coordinates along the z-axis of the 3D grid making the layers.
    layers: ndarray
        The values of the variable `var` based on a tie-centered scheme. This
        array has the same size as `x`, `y`, and `z` along its last two
        dimensions (the horizontal dimensions y and x), but one element less
        along its first dimension (the vertical dimension z).

    Reference
    ---------
    Tetzlaff, D. (2023)
        Stratigraphic forward modeling software package for research and education
        https://arxiv.org/abs/2302.05272
    """
    z = _get_layer_elevation(grid).reshape(-1, *grid.cell_grid_shape)
    y = grid.y_of_node[grid.core_nodes].reshape(grid.cell_grid_shape)
    y = np.tile(y, (len(z), 1, 1))
    x = grid.x_of_node[grid.core_nodes].reshape(grid.cell_grid_shape)
    x = np.tile(x, (len(z), 1, 1))
    _dz = _get_variable_values(grid, "dz")
    dz = _select_sublayers(_dz, np.sum, axis=2).reshape(-1, *grid.cell_grid_shape)
    if (
        isinstance(var, np.ndarray) == False
        and (var == "_dz" or var == "dz" or var == "thickness")
        and i_class == np.sum
    ):
        layers = dz
    else:
        layers = _select_sublayers(
            _get_variable_values(grid, var), i_class, axis=2, dz=_dz
        ).reshape(-1, *grid.cell_grid_shape)
    if mask_value is not None:
        layers[layers == mask_value] = np.nan

    for l in range(layers.shape[0]):
        for c in range(layers.shape[axis]):
            slices = tuple(
                slice(None) if i != axis else c for i in range(1, layers.ndim)
            )
            mask_layer(
                layers[l, *slices], dz[l, *slices], mask_wedges, mask_null_layers
            )
            if fill_nan == True and l > 0:
                layers[l, np.isnan(layers[l])] = layers[l - 1, np.isnan(layers[l])]

    return x, y, z, layers



"""Cython functions for plotting"""

# MIT License

# Copyright (c) 2025 Guillaume Rongier

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


cimport cython
from libc.math cimport NAN


################################################################################
# Functions


@cython.boundscheck(False)
@cython.wraparound(False)
cpdef void mask_layer(
    cython.floating [:] layer,
    const cython.floating [:] thickness,
    const bint mask_wedges,
    const bint mask_null_layers,
) noexcept nogil:
    """Masks the parts of a layer with zero thickness."""
    cdef unsigned int i

    for i in range(len(thickness)):
        if thickness[i] == 0.:
            if mask_wedges == True and i > 0 and thickness[i - 1] > 0.:
                layer[i] = layer[i - 1]
            elif mask_wedges == True and i < len(thickness) - 1 and thickness[i + 1] > 0.:
                layer[i] = layer[i + 1]
            elif (mask_null_layers == True
                  and ((i > 0 and thickness[i - 1] == 0. and i < len(thickness) - 1 and thickness[i + 1] == 0.)
                       or (i == 0 and thickness[i + 1] == 0.)
                       or (i == len(thickness) - 1 and thickness[i - 1] == 0.))):
                layer[i] = NAN
